class TaggedSentence:
    def __init__(self):
        self.words = []  # OBTResult
        self.valence = 1

    def __iter__(self):
        return iter(self.words)

    def __str__(self):
        return ' '.join([str(word) for word in self.words])


class OBTResult:
    def __init__(self, original_word, list=list(), is_bank=False, bank_target=None, targeted_aspect=None):
        self.original_word = original_word  # String
        self.list = list  # List<OBTWord>
        self.sentiment_hits = set()  # SentiWordNetEntry
        self.is_bank = is_bank # boolean
        self.bank_target = bank_target  # OBTResult
        self.aspect = targeted_aspect  # BankEntityAspect

    def __str__(self) -> str:
        return self.original_word

    def __ne__(self, o) -> bool:
        return not self == o

    def __hash__(self) -> int:
        return hash(self.original_word + str(self.list))

    def __eq__(self, o) -> bool:
        return isinstance(o, self.__class__) and self.original_word == o.original_word and self.list == o.list


class OBTWord:
    def __init__(self, word, pos, morph_info=list()):
        self.word = word
        self.pos = pos
        self.morph_info = morph_info

    def __hash__(self) -> int:
        return hash(self.word + self.pos)

    def __str__(self) -> str:
        return self.word

    def __ne__(self, other) -> bool:
        return not self == other

    def __eq__(self, other) -> bool:
        return isinstance(other, self.__class__) and self.word == other.word and self.pos == other.pos


class SentiWordNetEntry:
    def __init__(self, id, part_of_speech, pos_score, neg_score, lemmas, explaination):
        self.sentiwordnet_id = id
        self.part_of_speech = part_of_speech
        self.positivity_score = float(pos_score)
        self.negativity_score = float(neg_score)
        self.objectivity_score = 1 - (self.positivity_score + self.negativity_score)
        self.lemmas = lemmas
        self.explanation = explaination

    def __hash__(self) -> int:
        return hash(self.sentiwordnet_id + self.part_of_speech)

    def __eq__(self, other):
        return isinstance(other,
                          self.__class__) and self.sentiwordnet_id == other.sentiwordnet_id and self.part_of_speech == other.part_of_speech

    def __ne__(self, other):
        return not self == other

    def __str__(self) -> str:
        return '{}; positive: {}, negative: {}, lemmas: {}; {}'.format(self.__class__,
                                                                       self.positivity_score,
                                                                       self.negativity_score,
                                                                       self.lemmas,
                                                                       self.explanation)


class NorskOrdvevEntry:
    def __init__(self, synset_id, lemmas) -> None:
        self.synset_id = synset_id
        self.lemmas = lemmas

    def __hash__(self) -> int:
        return hash(self.synset_id)

    def __eq__(self, other) -> bool:
        return isinstance(other, self.__class__) and self.synset_id == other.synset_id

    def __ne__(self, o) -> bool:
        return not self == o

    def __str__(self) -> str:
        return '{}-object synset_id: {}, lemmas: {}'.format(self.__class__,
                                                            self.synset_id,
                                                            self.lemmas)


class BankEntitetsAspekt:
    def __init__(self, parent, splitted_lines: list, level=0) -> None:
        self.level = level
        self.parent = parent
        splitted_lines = [line.replace('\n', '') for line in splitted_lines]
        self.synonyms = splitted_lines[0].split('/')
        splitted_lines.pop(0)

        direct_children = [i for i in range(len(splitted_lines)) if splitted_lines[i].count('  ') == 1]
        new_line_split = [line[2:] for line in splitted_lines]
        children = []
        for index, next_index in zip(direct_children[0:-1], direct_children[1:]):
            children.append(BankEntitetsAspekt(self, new_line_split[index:next_index], level=level+1))
        if len(direct_children) != 0:
            children.append(BankEntitetsAspekt(self, new_line_split[direct_children[-1]:], level=level+1))
        self.children = tuple(children)

    def __str__(self) -> str:
        return str(self.synonyms)

    def name_all_decendants(self):
        name_dict = {}
        for child in self.children:
            for name, obj in child.name_all_decendants().items():
                if name not in name_dict.keys():
                    name_dict[name] = obj
                else:
                    name_dict[name].extend(obj)
        for synonym in self.synonyms:
            if synonym not in name_dict.keys():
                name_dict[synonym] = [self]
            else:
                name_dict[synonym].append(self)
        return name_dict

    def find_middles_untill_lowest_common_ancestor(self, other):
        this_middles = [self]
        other_middles = [other]
        while this_middles[-1].parent != other_middles[-1].parent:
            if this_middles[-1].level == max(this_middles[-1].level, other_middles[-1].level):
                this_middles.append(this_middles[-1].parent)
            else:
                other_middles.append(other_middles[-1].parent)
        return this_middles[1:], other_middles[1:]
