import numpy as np
from io import StringIO
from TextClasses import OBTResult, OBTWord
import random, string
import re

# vurder om adverb skal tolkes som adjektiv for bedre treff i SentiWordNet
obt_to_sentiwordnet_pos_mapping = {
    'subst': 'n',
    'verb': 'v',
    'adj': 'a',
    'adv': 'r',
}

norsk_ordvev_to_obt_pos_mapping = {
    'Noun': 'subst',
    'Verb': 'verb',
    'Adjective': 'adj',
}

REVIEW_SEP = ' @@@REVIEW_SEP@@@\n\n'
SENTENCE_SEP = ' @@@SENTENCE_SEP@@@ '


def parse_obt_results(tagged_text, entity_map):
    words = []
    tagged_stream = StringIO(tagged_text)
    for line in tagged_stream:
        if line.strip() == '':
            continue
        if line.strip() == '<word>|</word>':
            neste = next(tagged_stream)
            while not neste.startswith('\t'):
                neste = next(tagged_stream)
            continue
        if line.strip() == '<word>' + REVIEW_SEP.strip() + '</word>':
            neste = next(tagged_stream)
            while not neste.startswith('\t'):
                neste = next(tagged_stream)
            yield words
            words = []
        elif line.startswith('<word>') and line.strip().endswith('</word>'):
            current_word = line. \
                replace('</word>', ''). \
                replace('<word>', ''). \
                strip()
            if current_word in entity_map.keys():
                words.append(OBTResult(entity_map[current_word], [], is_bank=True))
            elif True in [mhash in current_word for mhash in entity_map.keys()]:
                curr_hash = [mhash for mhash in entity_map.keys() if mhash in current_word][0]
                words.append(OBTResult(current_word.replace(curr_hash, entity_map[curr_hash]), []))
            else:
                words.append(OBTResult(current_word, []))
            neste = next(tagged_stream)
            while neste.strip() == '':
                neste = next(tagged_stream)
        else:
            if not line.startswith('\t'):
                print(tagged_text)
                raise RuntimeError('Error in obt-result. Should be \\t. Was ' + str(line))
            splitted = line.strip().split('"')
            rest = ' '.join(splitted[2:]).strip().split()
            if current_word in entity_map.keys() or (len(rest) > 1 and rest[0] == 'subst' and rest[1] == 'prop'):
                pos = 'np'
            else:
                if len(rest) == 0:
                    continue
                pos = rest[0]
            words[-1].list.append(OBTWord(word=splitted[1], pos=pos, morph_info=line[1:]))
    yield words


def parse_norsk_ordbank_same_way_as_obt(norsk_ordbank, text, entity_map):
    words = []
    for word in text.split():
        word = word.strip('.,:;?!-\'\"/\\(){}[]<>')
        words.append(OBTResult(word, norsk_ordbank.get(word, [])))
    return words


def count_negations(text, negation_words):
    return sum([text.count(neg) for neg in negation_words])


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def random_string(length):
    return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(length))


entity_mapping = {}
sorted_all_names = []


def init_mapping(all_names):
    global sorted_all_names, entity_mapping
    entity_mapping.clear()
    for unique_name in set(name[1] for name in all_names):
        entity_mapping[unique_name] = '@' + random_string(10)
    sorted_all_names = sorted(all_names, key=lambda x: len(x[0].split()), reverse=True)


'''
    :param raw_text: string
    :param all_names: list of (name_form, unique_name)
'''


def extract_entities(raw_text):
    if not len(entity_mapping) or not len(sorted_all_names):
        raise NotImplementedError("Must call init_mapping before extract_entities")
    used_entities = {}
    for name, unique_name in sorted_all_names:
        if name.lower() in raw_text.lower():
            raw_text = re.sub(re.compile(re.escape(name), re.IGNORECASE), entity_mapping.get(unique_name), raw_text)
            used_entities[entity_mapping[unique_name]] = unique_name
    return raw_text, used_entities


def flatten(liste, to_type=None):
    if to_type:
        return to_type([subitem for sublist in liste for subitem in sublist])
    return type(liste)([subitem for sublist in liste for subitem in sublist])


def collapse_conf_matrices(matrices):
    initial = {
            -1: {-1: 0, 0: 0, 1: 0},
            0: {-1: 0, 0: 0, 1: 0},
            1: {-1: 0, 0: 0, 1: 0}
    }
    for matrix in matrices.values():
        for i in range(-1, 2):
            for j in range(-1, 2):
                initial[i][j] += matrix[i][j]
    return initial

def conf_matrix_accuracy(matrix):
    return float(sum([matrix[i][i] for i in range(-1, 2)])) / sum([matrix[i][j] for i in range(-1, 2) for j in range(-1, 2)]) or 1

def conf_matrix_recall(matrix, true_class):
    t = float(matrix[true_class][true_class])
    d = float(sum([matrix[true_class][i] for i in range(-1, 2)]))
    if t == d == 0.0:
        return 1.0
    elif t == 0.0:
        return 0.0
    return t / d

def conf_matrix_precision(matrix, predicted_class):
    t = float(matrix[predicted_class][predicted_class])
    d = float(sum([matrix[i][predicted_class] for i in range(-1, 2)]))
    if t == d == 0.0:
        return 1.0
    elif t == 0.0:
        return 0.0
    return t / d

def conf_matrix_f_score(matrix, a_class):
    recall = conf_matrix_recall(matrix, a_class)
    presicion = conf_matrix_precision(matrix, a_class)
    if recall == presicion == 0.0:
        return 0.0
    return 2 * recall * presicion / (recall + presicion)

def print_conf_matrix(conf_matrix):
    print('\n' * 2)
    print('Real\tPredicted value')
    print('\t' + '\t'.join([str(v) for v in range(-1, 2)]))
    for real in range(-1, 2):
        print(real, end='\t')
        for pred in range(-1, 2):
            print(conf_matrix[real][pred], end='\t')
        print()

def print_top10(vectorizer, clf, class_labels):
    """Prints features with the highest coefficient values, per class"""
    feature_names = vectorizer.get_feature_names()
    for i, class_label in enumerate(class_labels):
        top10 = np.argsort(clf.coef_[i])[-10:]
        print("%s: %s" % (class_label,
                          " ".join(feature_names[j] for j in top10)))
