import argparse

# Feature set categories
FEATURE_SET_POS_FREQ = 'pos'
FEATURE_SET_BANK_LEXICON = 'bank_lexicon'
FEATURE_SET_EMOJIS = 'emojis'
FEATURE_SET_TERMINAL_MARKS = 'terminal_marks'
FEATURE_SET_LENGTH = 'text_lengths'
FEATURE_SET_TEXT_ANALYSIS = 'sentiment_lookup'

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Perform sentiment analysis on bank reviews')
    
    parser.add_argument('-p', '--pos', dest='feature_classes', action='append_const', const=FEATURE_SET_POS_FREQ, help='include part-of-speech frequencies in featureset')
    parser.add_argument('-b', '--bank', dest='feature_classes', action='append_const', const=FEATURE_SET_BANK_LEXICON, help='include custom bank lexicon in featureset')
    parser.add_argument('-e', '--emoji', dest='feature_classes', action='append_const', const=FEATURE_SET_EMOJIS, help='include emojis in featureset')
    parser.add_argument('-t', '--term', dest='feature_classes', action='append_const', const=FEATURE_SET_TERMINAL_MARKS, help='include terminal signs in featureset')
    parser.add_argument('-l', '--length', dest='feature_classes', action='append_const', const=FEATURE_SET_LENGTH, help='include text legnth based features in featureset')
    parser.add_argument('-s', '--senti', dest='feature_classes', action='append_const', const=FEATURE_SET_TEXT_ANALYSIS, help='include sentiment lookup in SWN in featureset')
    parser.add_argument('--all-permutations', dest='all_permutations', action='store_true', help='whether to create all possible permutations of the feature classes')
    parser.add_argument('-d', '--debug', dest='debug', action='store_true', help='whether to connect to a debug server')
    parser.add_argument('--filter-banks', dest='filter_banks', action='store_true', help='whether to filter sentiments towards the current bank')
    parser.add_argument('--propagate-aspect-sentiments', dest='propagate_aspects', action='store_true', help='whether to propagate the sentiments of aspects up in the aspect tree')
    parser.add_argument('--aspect-features', dest='aspect_features', action='store_true', help='whether to include sentiment values of each aspect in feature set')
    parser.add_argument('--show-conf-matrix', dest='show_conf_matrix', action='store_true', help='whether show the confusion matrices for each model')
    parser.add_argument('--show-other-measures', dest='show_other_measures', action='store_true', help='whether to show precision, recall and f-measure')

    args = parser.parse_args()
    print(args)

    if not args.feature_classes:
        parser.error("must define at least one feature class!")

    if args.debug:
        import pydevd

        pydevd.settrace('10.0.0.75', port=7853, stdoutToServer=True, stderrToServer=True, suspend=False)

    import sentiment_analysis

    features = sentiment_analysis.extract_features_from_dataset(
        feature_classes=args.feature_classes,
        filter_bank=args.filter_banks,
        all_aspects_as_features=args.aspect_features,
        propagate_aspects_up_tree=args.propagate_aspects
    )

    sentiment_analysis.train_and_evaluate(
        features,
        feature_classes=tuple(args.feature_classes),
        show_conf_matrix=args.show_conf_matrix,
        show_other_measures=args.show_other_measures,
        all_permutations=args.all_permutations
    )
