import re
from nltk.stem import WordNetLemmatizer
import time
import pickle
import os
from google.cloud import translate

try:
    from progressbar import ProgressBar
except ImportError:
    pass

translations_cache = {
    'no': {'en': {}},
    'en': {'no': {}}
}

if os.path.isfile('translationDump.pickle'):
    with open('translationDump.pickle', 'rb') as f:
        translations_cache.update(pickle.load(f))

# ensure translations are singular
wordnet_lemmatizer = WordNetLemmatizer()
wordnet_lemmatizer.lemmatize("hi")

translation_client = translate.Client.from_service_account_json('service_key.json')


def getTranslation(words, original_lang, target_lang):
    translations = [None] * len(words)
    if original_lang not in translations_cache.keys():
        raise ValueError("Original_lang is invalid")
    for i in range(len(words)):
        if words[i].lower() in translations_cache[original_lang][target_lang].keys():
            translations[i] = translations_cache[original_lang][target_lang][words[i].lower()]

    indexes_not_translated = [i for i in range(len(translations)) if translations[i] is None]

    # what does google translate say?
    if len(indexes_not_translated) != 0:
        result = translation_client.translate([words[i] for i in indexes_not_translated],
                                              format_='text',
                                              target_language=target_lang,
                                              source_language=original_lang)

    # store in cache and update translations
    for normal_index, not_translated_index in enumerate(indexes_not_translated):
        the_words = ' '.join(
            [wordnet_lemmatizer.lemmatize(w)
             for w in re.sub('\u200B', '', result[normal_index]['translatedText'].lower()).split(' ')])

        translations_cache[original_lang][target_lang][words[not_translated_index]] = the_words
        translations[not_translated_index] = the_words

    return translations


def findMostLikelySentiwordnetEquivalentToLemmalist(translated_lemmalist, most_common_pos,
                                                    sentiwordnet_with_lemma_key):
    correct_pos = set()
    all_pos = []
    for lemma in translated_lemmalist:
        if lemma in sentiwordnet_with_lemma_key.keys():
            for sentiobj in sentiwordnet_with_lemma_key[lemma]:
                # get only the most popular occurence for each (lemma, pos) pair
                if sentiobj.part_of_speech == most_common_pos or (sentiobj.part_of_speech == 'r' and most_common_pos == 'a'):
                    correct_pos.add(sentiobj)
                # all_pos.append(sentiobj)

    # correct_pos_best_matches = sorted(correct_pos, key=lambda lemma, sentiobj: numberOfLemmaInCommon(translated_lemmalist, sentiobj))
    return min(correct_pos, default=None, key=lambda sentiobj: avg([int(sense) for lemma, sense in sentiobj.lemmas if lemma in translated_lemmalist]))

def avg(li):
    return float(sum(li)) / len(li) or 1


def numberOfLemmaInCommon(lemmalist, sentiobj):
    return len(set(lemmalist).intersection([lemma for lemma, sense in sentiobj.lemmas]))

def storeTranslations():
    with open('translationDump.pickle', 'wb') as f:
        pickle.dump(translations_cache, f)
        print('Stored', len(translations_cache['no']['en']), 'translations')


def compileLookup(norwegian_ordvev_with_id_key, norwegian_ordvev_with_lemma_pos_key, sentiwordnet_with_id_pos_key,
                  sentiwordnet_with_lemma_key):
    a = time.time()
    import helpers
    with open('compiledNorwegianToPWNLookup.tab', 'w', encoding='utf8') as f:
        f.write('# Norwegian_id SentiWordNet_id\n')
        print('Press Enter to stop')
        bar = ProgressBar(max_value=len(norwegian_ordvev_with_id_key))
        for i, (id, lemmaposlist) in enumerate(norwegian_ordvev_with_id_key.items()):
            if len(lemmaposlist) == 1 and lemmaposlist[0][1] == 'subst' and lemmaposlist[0][0][0].isupper():
                # do not translate proper nouns
                best_match = None
            else:
                lemmalist, poslist = [], {}
                for lemma, pos in lemmaposlist:
                    lemmalist.append(lemma)
                    poslist[pos] = poslist.get(pos, 0) + 1
                lemmalist = list(set(lemmalist))
                most_common_pos = max(poslist, key=poslist.get)

                translations = getTranslation(lemmalist, 'no', 'en')
                best_match = findMostLikelySentiwordnetEquivalentToLemmalist(translations, helpers.obt_to_sentiwordnet_pos_mapping.get(most_common_pos, most_common_pos),
                                                                             sentiwordnet_with_lemma_key)
            if best_match is not None:
                f.write("{}\t{}-{}\n".format(id, best_match.sentiwordnet_id, best_match.part_of_speech))
            bar.update(i)
            if not keep_running:
                storeTranslations()
                exit(0)

    storeTranslations()
    print('Finished in', time.time() - a, 'seconds')


keep_running = True


def stopRunningListener():
    input()
    global keep_running
    keep_running = False


def readCompiledNoToPWN():
    lookup = {}
    with open('LexicalResources/compiledNorwegianToPWNLookup.tab', encoding='utf8') as f:
        for line in f:
            if line.strip().startswith("#"):
                continue
            l = line.split()
            lookup[l[0]] = tuple(l[1].split('-'))
    return lookup


if __name__ == '__main__':
    import readNorskOrdvev
    import readSentiWordNet
    from threading import Thread

    Thread(target=stopRunningListener).start()

    compileLookup(readNorskOrdvev.read_compiled_norsk_ordvev('id'),
                  readNorskOrdvev.read_compiled_norsk_ordvev(),
                  readSentiWordNet.loadWithIdPosKey('LexicalResources/SentiWordNet_3.0.0_20130122.txt'),
                  readSentiWordNet.loadWithLemmaKey('LexicalResources/SentiWordNet_3.0.0_20130122.txt'))
