import os
import time
import warnings
from itertools import combinations

import nltk.data
import yaml
from nltk.classify.scikitlearn import SklearnClassifier
from nltk.classify.util import accuracy
from sklearn.linear_model import LogisticRegression as MaxEntClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.svm import LinearSVC

import helpers
import norwegianPWNMapping
import norwegian_words
import readNorwegianWordnet
import readSentiWordNet
from TextClasses import *
from obt_tagger import OBTagger
from readNorskOrdvev import read_compiled_norsk_ordvev
from main import FEATURE_SET_POS_FREQ, FEATURE_SET_BANK_LEXICON, FEATURE_SET_EMOJIS, FEATURE_SET_TERMINAL_MARKS, FEATURE_SET_LENGTH, FEATURE_SET_TEXT_ANALYSIS

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from nltk.sentiment.util import SAD, HAPPY

with open('bank_structure_tree.txt', encoding='utf8') as f:
    bank_aspect_root = BankEntitetsAspekt(None, f.readlines())
named_bank_aspects = bank_aspect_root.name_all_decendants()
all_bank_aspects = helpers.flatten(named_bank_aspects.values(), to_type=set)

# try:
# start = time.time()
obt_tagger = OBTagger(os.path.join('../The-Oslo-Bergen-Tagger'))
# oslo_bergen_available = True
# except ImportError:
#
#     print('Oslo-Bergen-tagger not available')
#     oslo_bergen_available = False
#     start = time.time()
#
#     import readNorskOrdbank
#     norsk_ordbank = readNorskOrdbank.load()

# with open('APtaggerAnders.pickle', 'rb') as f:
#     tagger = pickle.load(f)

with open('LexicalResources/custom_bank_lexicon.yaml', encoding='utf8') as f:
    bank_lexicon = yaml.load(f)['words']

nor_tokenizer = nltk.data.load('tokenizers/punkt/norwegian.pickle')
word_net = readNorwegianWordnet.load_norwegian_wordnet()
sentiwordnet = readSentiWordNet.loadWithIdPosKey('LexicalResources/SentiWordNet_3.0.0_20130122.txt')
sentiwordnet_lemmakey = readSentiWordNet.loadWithLemmaKey('LexicalResources/SentiWordNet_3.0.0_20130122.txt')
norsk_ordvev_lemma_key = read_compiled_norsk_ordvev(key='lemma,pos')
norsk_ordvev_id_key = read_compiled_norsk_ordvev(key='id')
compiledNorwegianToPWNLookupList = norwegianPWNMapping.readCompiledNoToPWN()

# Features
KEY_NUM_WORDS = 'numWords'
KEY_AVG_SENT_LENGTH = 'avgSentLength'
KEY_AVG_WORD_LENGTH = 'avgWordLength'
KEY_NUM_NEGATIONS = 'numNegations'
KEY_NUM_EXCLAMATION_MARKS = 'numExclamation'
KEY_NUM_QUESTION_MARKS = 'numQuestions'
KEY_NUM_POSITIVE_EMOTICONS = 'numPosEmoticons'
KEY_NUM_NEGATIVE_EMOTICONS = 'numNegEmoticons'
KEY_NUM_NEUTRAL_EMOTICONS = 'numNeuEmoticons'

# Feature classes
all_feature_classes = (
    FEATURE_SET_POS_FREQ,
    FEATURE_SET_BANK_LEXICON,
    FEATURE_SET_EMOJIS,
    FEATURE_SET_TERMINAL_MARKS,
    FEATURE_SET_LENGTH,
    FEATURE_SET_TEXT_ANALYSIS
)

# Names in reviews.yaml
REVIEW_TEXT = 'text'
REVIEW_POLARITY = 'polarity'
REVIEW_ORIGIN = 'origin'
REVIEW_SUBJECTIVITY = 'subj'
REVIEW_TAGGED = 'tagged'



category_mapping = {
    KEY_NUM_WORDS: FEATURE_SET_LENGTH,
    KEY_AVG_SENT_LENGTH: FEATURE_SET_LENGTH,
    KEY_AVG_WORD_LENGTH: FEATURE_SET_LENGTH,
    KEY_NUM_NEGATIONS: FEATURE_SET_LENGTH,
    KEY_NUM_EXCLAMATION_MARKS: FEATURE_SET_TERMINAL_MARKS,
    KEY_NUM_QUESTION_MARKS: FEATURE_SET_TERMINAL_MARKS,
    KEY_NUM_POSITIVE_EMOTICONS: FEATURE_SET_EMOJIS,
    KEY_NUM_NEGATIVE_EMOTICONS: FEATURE_SET_EMOJIS,
    KEY_NUM_NEUTRAL_EMOTICONS: FEATURE_SET_EMOJIS,
    'adj': FEATURE_SET_POS_FREQ, 'adv': FEATURE_SET_POS_FREQ, 'det': FEATURE_SET_POS_FREQ,
    'inf-merke': FEATURE_SET_POS_FREQ,
    'interj': FEATURE_SET_POS_FREQ, 'konj': FEATURE_SET_POS_FREQ, 'prep': FEATURE_SET_POS_FREQ,
    'pron': FEATURE_SET_POS_FREQ,
    'sbu': FEATURE_SET_POS_FREQ, 'subst': FEATURE_SET_POS_FREQ, 'ukjent': FEATURE_SET_POS_FREQ,
    'verb': FEATURE_SET_POS_FREQ
}
for word in bank_lexicon:
    category_mapping['bank_lexicon_' + word] = FEATURE_SET_BANK_LEXICON
for aspect in all_bank_aspects:
    category_mapping[str(aspect) + '_positive'] = FEATURE_SET_TEXT_ANALYSIS
    category_mapping[str(aspect) + '_negative'] = FEATURE_SET_TEXT_ANALYSIS


def __load_reviews__(filename):
    with open(filename, encoding='utf8') as f:
        reviews = yaml.load(f)
    return reviews


def __extract_features(review, company: str, feature_set_categories: tuple,
                       filter_bank=True,
                       all_aspects_as_features=True,
                       propagate_aspects_up_tree=True):
    feature_set = {}

    raw_text = review[REVIEW_TEXT]
    text_with_coded_entities, entity_mapping = helpers.extract_entities(raw_text)

    # if oslo_bergen_available:
    #     # noinspection PyUnboundLocalVariable
    words_tagged = review[REVIEW_TAGGED]
    # if words_tagged != wold:
    #     print()
    # else:
    #     # noinspection PyUnboundLocalVariable
    #     words_tagged = helpers.parse_norsk_ordbank_same_way_as_obt(norsk_ordbank, raw_text, entity_mapping)

    # http://www.kunnskap.no:443/data/Courses/norskpluss_bokmaal/grammar/nekting.html


    # FEATURE: count number of each POS:
    if FEATURE_SET_POS_FREQ in feature_set_categories:
        pos = {
            'adj': 0, 'adv': 0, 'det': 0, 'inf-merke': 0,
            'interj': 0, 'konj': 0, 'prep': 0, 'pron': 0,
            'sbu': 0, 'subst': 0, 'ukjent': 0, 'verb': 0
        }
        for obt_result in words_tagged:
            for word in obt_result.list:
                if word.pos in pos.keys():
                    pos[word.pos] += 1
        feature_set.update(pos)

    if FEATURE_SET_LENGTH in feature_set_categories:
        # FEATURE: number of words
        feature_set[KEY_NUM_WORDS] = float(len(raw_text.split()))

        # FEATURE: average number of words in sentence
        sentences = nor_tokenizer.tokenize(raw_text)
        sent_lengths = [len(words.split()) for words in sentences]
        feature_set[KEY_AVG_SENT_LENGTH] = float(sum(sent_lengths)) / len(sent_lengths)

        # FEATURE: average word length
        all_words = [float(len(i.strip('.,:;?!-\'\"/\\(){}[]<>'))) for i in raw_text.split()]
        feature_set[KEY_AVG_WORD_LENGTH] = sum(all_words) / len(all_words)

        # FEATURE: number of negations in the text
        feature_set[KEY_NUM_NEGATIONS] = helpers.count_negations(raw_text,
                                                                 [neg for neg in norwegian_words.VALENCE_SHIFTERS.keys()
                                                                  if norwegian_words.VALENCE_SHIFTERS[neg] < 0])

    if FEATURE_SET_TERMINAL_MARKS in feature_set_categories:
        feature_set[KEY_NUM_EXCLAMATION_MARKS] = raw_text.count('!')
        feature_set[KEY_NUM_QUESTION_MARKS] = raw_text.count('?')

    if FEATURE_SET_EMOJIS in feature_set_categories:
        feature_set[KEY_NUM_POSITIVE_EMOTICONS] = sum([raw_text.count(emoticon) for emoticon in HAPPY])
        feature_set[KEY_NUM_NEGATIVE_EMOTICONS] = sum([raw_text.count(emoticon) for emoticon in SAD])

    if FEATURE_SET_BANK_LEXICON in feature_set_categories:
        for word in bank_lexicon:
            feature_set['bank_lexicon_' + word] = word in raw_text

    # Analyse sentiment with SentiWordNet
    if FEATURE_SET_TEXT_ANALYSIS in feature_set_categories:

        # divide into sentences and compute valence for each sentence
        tagged_sentences = []
        create_new = True
        # matches = 1
        for word in words_tagged:
            if create_new:
                tagged_sentences.append(TaggedSentence())
                create_new = False
            tagged_sentences[-1].words.append(word)
            if word.original_word in norwegian_words.VALENCE_SHIFTERS:
                # matches += 1
                tagged_sentences[-1].valence *= norwegian_words.VALENCE_SHIFTERS[word.original_word]
            if word.original_word in norwegian_words.SKILLETEGN:
                # tagged_sentences[-1].valence *= matches
                create_new = True
                # matches = 1

        # assign sentiment_target
        # First run, identify banks
        banks = [[] for _ in range(len(tagged_sentences))]
        for outer in range(len(tagged_sentences)):
            for inner in range(len(tagged_sentences[outer].words)):
                if tagged_sentences[outer].words[inner].is_bank:
                    banks[outer].append(inner)

        # Second run, assign sentiment_target to closest bank in sentence
        for tagged_sent_index in range(len(tagged_sentences)):
            for tagged_word_index in range(len(tagged_sentences[tagged_sent_index].words)):
                if len(banks[tagged_sent_index]):
                    tagged_sentences[tagged_sent_index].words[tagged_word_index].bank_target = \
                        tagged_sentences[tagged_sent_index].words[
                            min(banks[tagged_sent_index], key=lambda x: abs(tagged_word_index - x))]
                else:
                    tagged_sentences[tagged_sent_index].words[tagged_word_index].bank_target = \
                        OBTResult(company, is_bank=True)

        # identify aspects
        aspect_indexes = [[] for _ in range(len(tagged_sentences))]
        for outer in range(len(tagged_sentences)):
            for inner in range(len(tagged_sentences[outer].words)):
                for poss_word in tagged_sentences[outer].words[inner].list:
                    if poss_word.word in named_bank_aspects.keys():
                        tagged_sentences[outer].words[inner].aspect = named_bank_aspects[poss_word.word][0]
                        aspect_indexes[outer].append(inner)
                        break

        for sent_index in range(len(aspect_indexes)):

            for aspect_word in aspect_indexes[sent_index]:
                for obt_result in tagged_sentences[sent_index].words[aspect_word - 1::-1]:
                    if 'subst' in [word.pos for word in obt_result.list]:
                        break
                    obt_result.aspect = tagged_sentences[sent_index].words[aspect_word].aspect
                for obt_result in tagged_sentences[sent_index].words[aspect_word + 1:]:
                    if 'subst' in [word.pos for word in obt_result.list]:
                        break
                    obt_result.aspect = tagged_sentences[sent_index].words[aspect_word].aspect

        # if a match is discovered, add a tuple (word (lemma), sentiwordnet_id) to sentiment_hits
        # STEP 1: NorNet -> sentiwordnet
        for tagged_sent in tagged_sentences:
            for tagged_word in tagged_sent.words:
                for word in tagged_word.list:  # probably only one
                    for id in word_net.get((word.word, helpers.obt_to_sentiwordnet_pos_mapping.get(word.pos, word.pos)),
                                           []):
                        tagged_word.sentiment_hits.add(
                            sentiwordnet[(id, helpers.obt_to_sentiwordnet_pos_mapping[word.pos])])

        # STEP 2: Norsk Ordvev -> Google translate -> Find equivalent in SentiWordNet
        for tagged_sent in tagged_sentences:
            for tagged_word in [i for i in tagged_sent.words if
                                len(i.sentiment_hits) == 0]:  # filter out those we found in last step
                for word in tagged_word.list:  # probably only one
                    for id in norsk_ordvev_lemma_key.get((word.word, word.pos), []):
                        if id in compiledNorwegianToPWNLookupList.keys():
                            hit = compiledNorwegianToPWNLookupList[id]
                            hit = sentiwordnet[hit]
                            tagged_word.sentiment_hits.add(hit)

        positive_in_sentences = {k: [] for k in all_bank_aspects}
        negative_in_sentences = {k: [] for k in all_bank_aspects}

        logistic = lambda x: x #helpers.sigmoid

        for tagged_sent in tagged_sentences:
            for tagged_word in filter(
                    lambda obt_result: not filter_bank or obt_result.bank_target.original_word == company,
                    tagged_sent.words):
                tagged_word.aspect = tagged_word.aspect or bank_aspect_root
                for hit in tagged_word.sentiment_hits:
                    if hit.positivity_score:
                        if hit.positivity_score * tagged_sent.valence > 0:
                            positive_in_sentences[tagged_word.aspect].append(
                                logistic(hit.positivity_score * tagged_sent.valence) / len(tagged_word.sentiment_hits))
                        else:
                            negative_in_sentences[tagged_word.aspect].append(
                                logistic(abs(hit.positivity_score * tagged_sent.valence)) / len(tagged_word.sentiment_hits))
                    if hit.negativity_score:
                        if hit.negativity_score * tagged_sent.valence > 0:
                            negative_in_sentences[tagged_word.aspect].append(
                                logistic(hit.negativity_score * tagged_sent.valence) / len(tagged_word.sentiment_hits))
                        else:
                            positive_in_sentences[tagged_word.aspect].append(
                                logistic(abs(hit.negativity_score * tagged_sent.valence)) / len(tagged_word.sentiment_hits))

        max_depth = max(positive_in_sentences.keys(), key=lambda x: x.level).level
        aspect_levels = [[asp for asp in positive_in_sentences.keys() if asp.level == level] for level in
                         range(max_depth + 1)]

        if propagate_aspects_up_tree:
            # propagate sentiments, slice away root and go backwards
            for aspectset in aspect_levels[:0:-1]:
                for aspect in aspectset:
                    positive_in_sentences[aspect.parent].extend(positive_in_sentences[aspect])
                    negative_in_sentences[aspect.parent].extend(negative_in_sentences[aspect])

        if all_aspects_as_features:
            for aspect in all_bank_aspects:
                feature_set[str(aspect) + '_positive'] = sum(positive_in_sentences[aspect])
                feature_set[str(aspect) + '_negative'] = sum(negative_in_sentences[aspect])
        else:
            feature_set[str(bank_aspect_root) + '_positive'] = sum(positive_in_sentences[bank_aspect_root])
            feature_set[str(bank_aspect_root) + '_negative'] = sum(negative_in_sentences[bank_aspect_root])
    return feature_set


def extract_features_from_dataset(
        reviews_file='reviews.yaml',
        feature_classes=all_feature_classes,
        **kwargs
):
    reviews = __load_reviews__(reviews_file)

    all_names = sorted([(name, company) for company in reviews.keys() for name in reviews[company]['names']],
                       key=lambda x: len(x[0].split()), reverse=True)
    helpers.init_mapping(all_names)

    a = time.time()
    for company in reviews.keys():
        full_text = []
        for review in reviews[company]['reviews']:
            full_text.append(review[REVIEW_TEXT])
        full_text = helpers.REVIEW_SEP.join(full_text)
        text_with_coded_entities, entity_mapping = helpers.extract_entities(full_text)

        for review, tagged in zip(reviews[company]['reviews'],
                                  helpers.parse_obt_results(obt_tagger.run(text_with_coded_entities), entity_mapping)):
            review[REVIEW_TAGGED] = tagged

    print('Finished tagging reviews in', time.time() - a, 'seconds')
    features = {}
    for company in reviews.keys():
        features[company] = []
        for review in reviews[company]['reviews']:
            features[company].append((
                __extract_features(review, company, feature_classes, **kwargs),
                review[REVIEW_POLARITY]
            ))
    return features


MAXENT_DEFAULT_ARGS = {}
SVC_DEFAULT_ARGS = {}
NB_DEFAULT_ARGS = {}


def train_and_evaluate(
        features_grouped_on_company: dict,
        classifiers=(GaussianNB, LinearSVC, MaxEntClassifier),
        k_fold=5,
        feature_classes=all_feature_classes,
        combination_max_size=len(all_feature_classes),
        show_conf_matrix=False,
        show_other_measures=False,
        all_permutations=True
):
    # confusion matrix on form confusion_matrix{[company]}[true_value][predicted_value]

    if all_permutations:
        all_combinations = []
        for i in range(combination_max_size):
            all_combinations.extend(combinations(feature_classes, i + 1))
    else:
        all_combinations = [tuple(feature_classes)]

    featureset_accuracies = {}
    # print(end='\t')
    # for classifier in classifiers:
    #     print(classifier.__name__, end='\t')
    # print()
    for featureset in all_combinations:
        featureset_accuracies[featureset] = []
        # filter features
        current_features = {k: [] for k in features_grouped_on_company.keys()}
        for company, comp_feat_set in features_grouped_on_company.items():
            for original_features, polarity in comp_feat_set:
                current_features[company].append((
                    {k: v for k, v in original_features.items() if category_mapping[k] in featureset},
                    polarity
                ))
        # print(','.join(featureset), end='\t')
        for classifier in classifiers:
            confusion_matrices = {
                k: {
                    -1: {-1: 0, 0: 0, 1: 0},
                    0: {-1: 0, 0: 0, 1: 0},
                    1: {-1: 0, 0: 0, 1: 0}
                } for k in features_grouped_on_company.keys()
            }
            classifier_acc = []
            avg_acc = []
            # trainin = before + after
            # test = middle
            for fold in range(k_fold):
                # slice into training set and testing set in k-folds
                full_tr_set = []
                tes_set_comp = {}
                for company, features in current_features.items():
                    training_set, testing_set = \
                        features[:int(len(features) * float(fold) / k_fold)] + features[int(
                            len(features) * float(fold + 1) / k_fold):len(features)], \
                        features[int(len(features) * float(fold) / k_fold):int(
                            len(features) * float(fold + 1) / k_fold)]
                    full_tr_set.extend(training_set)
                    tes_set_comp[company] = testing_set
                # print('Create', classifier.__name__, 'classifiers')
                wrapper = SklearnClassifier(classifier(), sparse=False)

                # print('Start training...')
                # now = time.time()
                wrapper.train(full_tr_set)
                # print('Training finished in', time.time() - now, 'seconds')
                # print('Start testing...')
                # print('.Before:', wrapper._vectorizer.get_feature_names())
                # wrapper._vectorizer.restrict(SelectFromModel(wrapper._clf, prefit=True))
                # print('After:', wrapper._vectorizer.get_feature_names())
                for company, test_set in tes_set_comp.items():
                    results = wrapper.classify_many([fs for (fs, l) in test_set])
                    for ((feat, correct), predicted) in zip(test_set, results):
                        confusion_matrices[company][correct][predicted] += 1
                    # this_acc = accuracy(wrapper, testing_set)
                    # avg_acc.append(this_acc)
            #     classifier_acc.append(float(sum(avg_acc)) / len(avg_acc))
            # total_classifier_acc = float(sum(classifier_acc)) / len(classifier_acc)

            # print(total_classifier_acc, end='\t')
            # print('\n'*2)
            # print()

            tot_conf_matrix = helpers.collapse_conf_matrices(confusion_matrices)
            if show_other_measures or show_conf_matrix:
                print()
                print(classifier.__name__, 'classifier with feat_set', featureset)
            if show_conf_matrix:
                helpers.print_conf_matrix(tot_conf_matrix)
                print('Accuracy:', helpers.conf_matrix_accuracy(tot_conf_matrix))
            if show_other_measures:
                for cls_name, value in zip(("Negative", "Neutral", "Positive"), range(-1, 2)):
                    print(cls_name, end='\t')
                    print('Precision:', helpers.conf_matrix_precision(tot_conf_matrix, value), end='\t')
                    print('Recall:', helpers.conf_matrix_recall(tot_conf_matrix, value), end='\t')
                    print('F-score:', helpers.conf_matrix_f_score(tot_conf_matrix, value), end='\t')
            featureset_accuracies[featureset].append(helpers.conf_matrix_accuracy(tot_conf_matrix))

    print()
    print('\t', end='')
    for classifier in classifiers:
        print(classifier.__name__, end='\t')
    print()

    for feats, accuracies in sorted(featureset_accuracies.items(), key=lambda tup: max(tup[1]), reverse=True):
        print(feats, end='\t')
        for ac in accuracies:
            print(str(ac).replace('.', ','), end='\t')
        print()
